﻿using Force.Crc32;
using Hprose.RPC;
using PackageInterface;
using System;
using System.IO;
using System.IO.Compression;
using System.Threading;

namespace PackageClient
{
    class Program
    {
        static void Main(string[] args)
        {
            //Special cmd for killing the packager.
            if (args.Length == 3 && args[0] == "kill")
            {
                var projectName = args[1];
                var address = args[2];
                try
                {
                    var client = new Client(address) { Timeout = TimeSpan.FromMinutes(30) };
                    var packageManager = client.UseService<IPackageManager>(nameof(IPackageManager));

                    Console.WriteLine($"Getting or creating packager for project: {projectName}...");
                    var packagerId = packageManager.GetOrCreatePackager(projectName);
                    if (string.IsNullOrEmpty(packagerId))
                    {
                        throw new InvalidOperationException("There is no avaliable packager.");
                    }

                    Console.WriteLine($"Get packager: {packagerId}");

                    Console.WriteLine($"Killing packager: {packagerId}...");
                    packageManager.KillPackager(packagerId);
                    Console.WriteLine($"Packager {packagerId} was killed.");

                }
                catch(Exception ex)
                {
                    Console.WriteLine($"Kill packager failed, error:{ex}.");
                }
            }
            else
            {
                if (args.Length != 5)
                {
                    Console.WriteLine("Running arguments:");
                    Console.WriteLine("1. Project Name: string - The name of the project");
                    Console.WriteLine("2. Version: string - The version of package files");
                    Console.WriteLine("3. File Folder: string - The folder of the files to transfer.");
                    Console.WriteLine("4. Package: bool - Execute the package script.");
                    Console.WriteLine("5. Server Address: string - The server address.");
                    return;
                }
                try
                {
                    var projectName = args[0];
                    var version = args[1];
                    var folder = args[2];
                    var package = bool.Parse(args[3]);
                    var address = args[4];

                    var client = new Client(address) { Timeout = TimeSpan.FromMinutes(30) };
                    var packageManager = client.UseService<IPackageManager>(nameof(IPackageManager));

                    Console.WriteLine($"Getting or creating packager for project: {projectName}...");
                    var packagerId = packageManager.GetOrCreatePackager(projectName);
                    if (string.IsNullOrEmpty(packagerId))
                    {
                        throw new InvalidOperationException("There is no avaliable packager.");
                    }

                    Console.WriteLine($"Get packager: {packagerId}");

                    Console.WriteLine("Starting package task...");
                    var taskId = packageManager.StartPackageTask(packagerId, version);
                    Console.WriteLine($"Task {taskId} started.");

                    Console.WriteLine($"Creating transfer package from folder {folder}...");

                    var transferFolder = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "Transfer");
                    if(!Directory.Exists(transferFolder))
                    {
                        Directory.CreateDirectory(transferFolder);
                    }
                    var projectFolder = Path.Combine(transferFolder, projectName);
                    if(!Directory.Exists(projectFolder))
                    {
                        Directory.CreateDirectory(projectFolder);
                    }

                    var transferFile = Path.Combine(projectFolder, "package.zip");
                    if (File.Exists(transferFile))
                    {
                        File.Delete(transferFile);
                    }
                    ZipFile.CreateFromDirectory(folder, transferFile);

                    Console.WriteLine("Loading file data...");
                    var fileData = File.ReadAllBytes(transferFile);
                    Console.WriteLine($"Transfer file size: {fileData.Length}");
                    var hashCode = Crc32Algorithm.Compute(fileData);
                    Console.WriteLine($"Transfer file hash code: {hashCode}");

                    if (File.Exists(transferFile))
                    {
                        File.Delete(transferFile);
                    }

                    Console.WriteLine($"Start transfering...");
                    //Transfer
                    packageManager.StartTransfer(packagerId, taskId, hashCode);

                    var fileSize = fileData.Length;
                    var transferCount = fileSize / 1048576;
                    var lastBlockSize = fileSize % 1048576;

                    var totalCount = transferCount + (lastBlockSize > 0 ? 1 : 0);

                    for (var i = 0; i < transferCount; i++)
                    {
                        var transferData = new byte[1048576];
                        Array.Copy(fileData, i * 1048576, transferData, 0, transferData.Length);
                        packageManager.TransferData(packagerId, taskId, transferData);

                        var progress = (int)((float)i / totalCount * 100);
                        Console.WriteLine($"{progress}% transfered.");
                    }
                    if (lastBlockSize > 0)
                    {
                        var transferData = new byte[lastBlockSize];
                        Array.Copy(fileData, transferCount * 1048576, transferData, 0, transferData.Length);
                        packageManager.TransferData(packagerId, taskId, transferData);
                        Console.WriteLine($"100% transfered.");
                    }
                    packageManager.EndTransfer(packagerId, taskId);

                    if (package)
                    {
                        Console.WriteLine($"Start packaging...");
                        packageManager.Execute(packagerId, taskId, "package.bat");
                        var taskState = packageManager.GetPackageTaskState(packagerId, taskId);
                        while (taskState != "Running")
                        {
                            Thread.Sleep(1000);
                            taskState = packageManager.GetPackageTaskState(packagerId, taskId);
                        }
                        Console.WriteLine($"Package finished.");
                    }

                    packageManager.EndPackage(packagerId, taskId);
                    Console.WriteLine($"Package task end.");

                }
                catch (Exception ex)
                {
                    Console.WriteLine($"Running package client error:{ex}");
                }
            }
        }
    }
}
